@@ -15,6 +15,7 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <string.h>
+#include <stdio.h>
 
 #ifdef LINC_SSL_SUPPORT
 #include <openssl/ssl.h>
@@ -32,7 +33,13 @@ enum {
 	BLOCKING,
 	LAST_SIGNAL
 };
-static guint linc_connection_signals [LAST_SIGNAL];
+static guint   signals [LAST_SIGNAL];
+static GMutex *cnx_lock = NULL;
+static GCond  *cnx_condition = NULL;
+
+#define CNX_LOCK(cnx)      LINC_MUTEX_LOCK (cnx_lock)
+#define CNX_UNLOCK(cnx)    LINC_MUTEX_UNLOCK (cnx_lock)
+#define CNX_IS_LOCKED(cnx) linc_mutex_is_locked (cnx_lock)
 
 static gboolean linc_connection_io_handler (GIOChannel  *gioc,
 					    GIOCondition condition,
@@ -72,23 +79,21 @@ queue_signal (LINCConnection *cnx,
 	cnx->priv->write_queue_bytes += delta;
 	new_size = cnx->priv->write_queue_bytes;
 
-	g_object_ref (G_OBJECT (cnx));
+	linc_object_ref (cnx);
 
 	if (cnx->options & LINC_CONNECTION_BLOCK_SIGNAL) {
 		if (new_size == 0 ||
 		    (old_size < (cnx->priv->max_buffer_bytes >> 1) &&
 		     new_size >= (cnx->priv->max_buffer_bytes >> 1)) ||
 		    new_size >= cnx->priv->max_buffer_bytes)
-			g_signal_emit (G_OBJECT (cnx),
-				       linc_connection_signals [BLOCKING],
-				       0, new_size);
+			g_signal_emit (cnx, signals [BLOCKING], 0, new_size);
 	}
 
 	if (cnx->priv->max_buffer_bytes &&
 	    cnx->priv->write_queue_bytes >= cnx->priv->max_buffer_bytes)
 		linc_connection_state_changed (cnx, LINC_DISCONNECTED);
 
-	g_object_unref (G_OBJECT (cnx));
+	linc_object_unref (cnx);
 }
 
 static gulong
@@ -107,7 +112,8 @@ calc_size (struct iovec *src_vecs,
 static void
 queue_flattened (LINCConnection *cnx,
 		 struct iovec   *src_vecs,
-		 int             nvecs)
+		 int             nvecs,
+		 gboolean        update_poll)
 {
 	int     i;
 	guchar *p;
@@ -136,6 +142,19 @@ queue_flattened (LINCConnection *cnx,
 
 	cnx->priv->write_queue = g_list_append (cnx->priv->write_queue, qw);
 	queue_signal (cnx, total_size);
+
+
+	if (update_poll) {
+		LINCCommandSetCondition *cmd;
+
+		g_assert (linc_get_threaded ());
+
+		cmd = g_new (LINCCommandSetCondition, 1);
+		cmd->cmd.type = LINC_COMMAND_SET_CONDITION;
+		cmd->cnx = linc_object_ref (cnx);
+		cmd->condition = (LINC_ERR_CONDS | LINC_IN_CONDS | G_IO_OUT);
+		linc_exec_command (&cmd->cmd);
+	}
 }
 
 static void
@@ -242,7 +261,7 @@ linc_connection_class_state_changed (LINCConnection      *cnx,
 		/* don't free pending queue - we could get re-connected */
 		if (changed)
 			g_signal_emit (G_OBJECT (cnx),
-				       linc_connection_signals [BROKEN], 0);
+				       signals [BROKEN], 0);
 		break;
 	}
 }
@@ -261,6 +280,8 @@ linc_connection_class_state_changed (LINCConnection      *cnx,
  * Fill in @cnx, call protocol specific initialisation methonds and then
  * call linc_connection_state_changed.
  *
+ * Must be called from the I/O thread.
+ *
  * Return Value: #TRUE if the function succeeds, #FALSE otherwise.
  */
 gboolean
@@ -273,6 +294,8 @@ linc_connection_from_fd (LINCConnection         *cnx,
 			 LINCConnectionStatus    status,
 			 LINCConnectionOptions   options)
 {
+	CNX_LOCK (cnx);
+
 	cnx->was_initiated = was_initiated;
 	cnx->is_auth       = (proto->flags & LINC_PROTOCOL_SECURE);
 	cnx->proto         = proto;
@@ -296,9 +319,10 @@ linc_connection_from_fd (LINCConnection         *cnx,
 		SSL_set_fd (cnx->priv->ssl, fd);
 	}
 #endif
-
 	linc_connection_state_changed (cnx, status);
 
+	CNX_UNLOCK (cnx);
+
   	return TRUE;
 }
 
@@ -340,6 +364,13 @@ linc_connection_initiate (LINCConnection        *cnx,
 	saddr = linc_protocol_get_sockaddr (
 		proto, host, service, &saddr_len);
 
+	if (!saddr && (strcmp (proto_name, "IPv6") ==0)) {/* Falling back to IPv4 */
+		proto = linc_protocol_find ("IPv4");
+		
+		saddr = linc_protocol_get_sockaddr (
+			proto, host, service, &saddr_len);
+	}
+
 	if (!saddr)
 		return FALSE;
 
@@ -386,6 +417,9 @@ linc_connection_initiate (LINCConnection        *cnx,
  * @status: a #LINCConnectionStatus.
  *
  * A wrapper for the #LINCConnectionClass's state change method.
+ *
+ * NB. this method must be called only in the I/O thread, or
+ * at construction, and with the cnx lock held ...
  */
 void
 linc_connection_state_changed (LINCConnection      *cnx,
@@ -393,6 +427,8 @@ linc_connection_state_changed (LINCConnection      *cnx,
 {
 	LINCConnectionClass *klass;
 
+	g_assert (CNX_IS_LOCKED (cnx));
+
 	klass = (LINCConnectionClass *)G_OBJECT_GET_CLASS (cnx);
 
 	if (klass->state_changed)
@@ -407,6 +443,7 @@ linc_connection_state_changed (LINCConnection      *cnx,
  * @block_for_full_read: whether to block for a full read
  * 
  * Warning, block_for_full_read is of limited usefullness.
+ * Can only be used from the main thread / with the cnx locked.
  *
  * Return value: number of bytes written on success; negative on error.
  **/
@@ -418,13 +455,15 @@ linc_connection_read (LINCConnection *cnx,
 {
 	int bytes_read = 0;
 
+	g_assert (CNX_IS_LOCKED (cnx));
+
 	d_printf ("Read up to %d bytes from fd %d\n", len, cnx->priv->fd);
 
 	if (!len)
 		return 0;
 
 	if (cnx->status != LINC_CONNECTED)
-		return LINC_IO_FATAL_ERROR;
+		goto fatal_error;
 
 	do {
 		int n;
@@ -448,9 +487,9 @@ linc_connection_read (LINCConnection *cnx,
 				if ((rv == SSL_ERROR_WANT_READ ||
 				     rv == SSL_ERROR_WANT_WRITE) &&
 				    (cnx->options & LINC_CONNECTION_NONBLOCKING))
-					return bytes_read;
+					goto out;
 				else
-					return LINC_IO_FATAL_ERROR;
+					goto fatal_error;
 			} else
 #endif
 			{
@@ -459,19 +498,20 @@ linc_connection_read (LINCConnection *cnx,
 
 				else if (errno == EAGAIN &&
 					 (cnx->options & LINC_CONNECTION_NONBLOCKING))
-					return bytes_read;
+					goto out;
 
 				else if (errno == EBADF) {
 					g_warning ("Serious fd usage error %d", cnx->priv->fd);
-					return LINC_IO_FATAL_ERROR;
+					goto fatal_error;
 
 				} else
-					return LINC_IO_FATAL_ERROR;
+					goto fatal_error;
 			}
 
 		} else if (n == 0) {
 			d_printf ("we got EOF on fd %d\n", cnx->priv->fd);
-			return LINC_IO_FATAL_ERROR;
+			bytes_read = LINC_IO_FATAL_ERROR;
+			goto out;
 		} else {
 			buf += n;
 			len -= n;
@@ -481,7 +521,11 @@ linc_connection_read (LINCConnection *cnx,
 
 	d_printf ("we read %d bytes\n", bytes_read);
 
+ out:
 	return bytes_read;
+
+ fatal_error:
+	return LINC_IO_FATAL_ERROR;
 }
 
 /* Determine the maximum size of the iovec vector */
@@ -599,7 +643,7 @@ write_data (LINCConnection *cnx, QueuedWrite *qw)
 
 			if (n) {
 				qw->vecs->iov_len  -= n;
-				qw->vecs->iov_base += n;
+				qw->vecs->iov_base = (guchar *)qw->vecs->iov_base + n;
 			}
 		}
 	}
@@ -620,6 +664,69 @@ linc_connection_should_block (LINCConnection      *cnx,
 	return FALSE;
 }
 
+/* Always called in main thread */
+static void
+linc_connection_flush_write_queue (LINCConnection *cnx)
+{
+	gboolean done_writes = TRUE;
+
+	if (cnx->priv->write_queue) {
+		glong        status;
+		QueuedWrite *qw = cnx->priv->write_queue->data;
+
+		status = write_data (cnx, qw);
+
+		d_printf ("Wrote queue %ld on fd %d\n", status, cnx->priv->fd);
+
+		if (status >= LINC_IO_OK) {
+			cnx->priv->write_queue = g_list_delete_link
+				(cnx->priv->write_queue, cnx->priv->write_queue);
+			queued_write_free (qw);
+
+			queue_signal (cnx, -status);
+			
+			done_writes = (cnx->priv->write_queue == NULL);
+
+		} else {
+			if (status == LINC_IO_FATAL_ERROR) {
+				d_printf ("Fatal error on queued write");
+				linc_connection_state_changed (cnx, LINC_DISCONNECTED);
+				
+			} else {
+				d_printf ("Write blocked\n");
+				done_writes = FALSE;
+			}
+		}
+	}
+
+	d_printf ("Blocked write queue %s\n", done_writes ?
+		  "flushed & empty" : "still active");
+
+	if (done_writes) /* drop G_IO_OUT */
+		linc_watch_set_condition
+			(cnx->priv->tag,
+			 LINC_ERR_CONDS | LINC_IN_CONDS);
+	else
+		linc_watch_set_condition
+			(cnx->priv->tag,
+			 LINC_ERR_CONDS | LINC_IN_CONDS | G_IO_OUT);
+}
+
+void
+linc_connection_exec_set_condition (LINCCommandSetCondition *cmd)
+{
+	d_printf ("Exec defered set condition on %p -> 0x%x",
+		  cmd->cnx, cmd->condition);
+
+	CNX_LOCK (cmd->cnx);
+	linc_watch_set_condition
+		(cmd->cnx->priv->tag,
+		 LINC_ERR_CONDS | LINC_IN_CONDS | G_IO_OUT);
+	CNX_UNLOCK (cmd->cnx);
+	linc_object_unref (cmd->cnx);
+	g_free (cmd);
+}
+
 /**
  * linc_connection_writev:
  * @cnx: the connection to write to
@@ -644,19 +751,28 @@ linc_connection_writev (LINCConnection       *cnx,
 	QueuedWrite qw;
 	int         status;
 
-	/* FIXME: need an option to turn this off ? */
-	if (cnx->options & LINC_CONNECTION_NONBLOCKING) {
-		while (cnx->status == LINC_CONNECTING)
-			linc_main_iteration (TRUE);
-	}
+	CNX_LOCK (cnx);
 
-	g_return_val_if_fail (cnx->status == LINC_CONNECTED,
-			      LINC_IO_FATAL_ERROR);
+	if (linc_get_threaded ()) {
+		d_printf ("Transfer output to main thread\n");
+		if (cnx->status == LINC_CONNECTING) {
+			queue_flattened (cnx, vecs, nvecs, TRUE);
+			CNX_UNLOCK (cnx);
+			return LINC_IO_QUEUED_DATA;
+		}
+	} else if (cnx->options & LINC_CONNECTION_NONBLOCKING)
+		linc_connection_wait_connected (cnx);
+
+	if (cnx->status == LINC_DISCONNECTED) {
+		CNX_UNLOCK (cnx);
+		return LINC_IO_FATAL_ERROR;
+	}
 
 	if (cnx->priv->write_queue) {
 		/* FIXME: we should really retry the write here, but we'll
 		 * get a POLLOUT for this lot at some stage anyway */
-		queue_flattened (cnx, vecs, nvecs);
+		queue_flattened (cnx, vecs, nvecs, FALSE);
+		CNX_UNLOCK (cnx);
 		return LINC_IO_QUEUED_DATA;
 	}
 
@@ -667,12 +783,20 @@ linc_connection_writev (LINCConnection       *cnx,
 	status = write_data (cnx, &qw);
 
 	if (status == LINC_IO_QUEUED_DATA) {
+		if (linc_get_threaded ()) {
+			queue_flattened (cnx, qw.vecs, qw.nvecs, TRUE);
+			CNX_UNLOCK (cnx);
+			return LINC_IO_QUEUED_DATA;
+		}
+
 		/* Queue data & listen for buffer space */
-		linc_watch_set_condition (cnx->priv->tag,
-					  LINC_ERR_CONDS | LINC_IN_CONDS | G_IO_OUT);
+		linc_watch_set_condition
+			(cnx->priv->tag,
+			 LINC_ERR_CONDS | LINC_IN_CONDS | G_IO_OUT);
 
 		if (!linc_connection_should_block (cnx, opt_write_opts)) {
-			queue_flattened (cnx, qw.vecs, qw.nvecs);
+			queue_flattened (cnx, qw.vecs, qw.nvecs, FALSE);
+			CNX_UNLOCK (cnx);
 			return LINC_IO_QUEUED_DATA;
 
 		} else {
@@ -683,6 +807,8 @@ linc_connection_writev (LINCConnection       *cnx,
 	} else if (status >= LINC_IO_OK)
 		status = LINC_IO_OK;
 
+	CNX_UNLOCK (cnx);
+
 	return status;
 }
 
@@ -759,13 +885,17 @@ linc_connection_class_init (LINCConnectionClass *klass)
 {
 	GObjectClass *object_class = (GObjectClass *) klass;
 
+	cnx_lock = linc_mutex_new();
+	if (cnx_lock)
+		cnx_condition = g_cond_new();
+
 	object_class->dispose  = linc_connection_dispose;
 	object_class->finalize = linc_connection_finalize;
 
 	klass->state_changed  = linc_connection_class_state_changed;
 	klass->broken         = NULL;
 
-	linc_connection_signals [BROKEN] =
+	signals [BROKEN] =
 		g_signal_new ("broken",
 			      G_TYPE_FROM_CLASS (object_class),
 			      G_SIGNAL_RUN_LAST,
@@ -774,7 +904,7 @@ linc_connection_class_init (LINCConnectionClass *klass)
 			      g_cclosure_marshal_VOID__VOID,
 			      G_TYPE_NONE, 0);
 
-	linc_connection_signals [BLOCKING] =
+	signals [BLOCKING] =
 		g_signal_new ("blocking",
 			      G_TYPE_FROM_CLASS (object_class),
 			      G_SIGNAL_RUN_LAST,
@@ -836,8 +966,11 @@ linc_connection_set_max_buffer (LINCConnection *cnx,
 {
 	g_return_if_fail (cnx != NULL);
 
+	CNX_LOCK (cnx);
 	/* FIXME: we might want to check the current buffer size */
 	cnx->priv->max_buffer_bytes = max_buffer_bytes;
+
+	CNX_UNLOCK (cnx);
 }
 
 static gboolean
@@ -847,10 +980,10 @@ linc_connection_io_handler (GIOChannel  *gioc,
 {
 	LINCConnection      *cnx = data;
 	LINCConnectionClass *klass;
-	int rv, n;
-	LincSockLen n_size = sizeof (n);
 
-	g_object_ref (G_OBJECT (cnx));
+	linc_object_ref (cnx);
+
+	CNX_LOCK (cnx);
 
 	klass = (LINCConnectionClass *) G_TYPE_INSTANCE_GET_CLASS (
 		data, LINC_TYPE_CONNECTION, LINCConnection);
@@ -859,52 +992,19 @@ linc_connection_io_handler (GIOChannel  *gioc,
 	    condition & LINC_IN_CONDS && klass->handle_input) {
 		
 		d_printf ("Handle input on fd %d\n", cnx->priv->fd);
-		klass->handle_input (cnx);
 
+		klass->handle_input (cnx);
 	}
 
 	if (cnx->status == LINC_CONNECTED && condition & G_IO_OUT) {
-		gboolean done_writes = TRUE;
-
 		d_printf ("IO Out - buffer space free ...\n");
-
-		if (cnx->priv->write_queue) {
-			glong        status;
-			QueuedWrite *qw = cnx->priv->write_queue->data;
-
-			status = write_data (cnx, qw);
-
-			d_printf ("Wrote queue %ld on fd %d\n", status, cnx->priv->fd);
-
-			if (status >= LINC_IO_OK) {
-				cnx->priv->write_queue = g_list_delete_link (
-					cnx->priv->write_queue, cnx->priv->write_queue);
-				queued_write_free (qw);
-
-				queue_signal (cnx, -status);
-
-				done_writes = (cnx->priv->write_queue == NULL);
-
-			} else if (status == LINC_IO_FATAL_ERROR) {
-				d_printf ("Fatal error on queued write");
-				linc_connection_state_changed (cnx, LINC_DISCONNECTED);
-
-			} else {
-				d_printf ("Write blocked\n");
-				done_writes = FALSE;
-			}
-		}
-
-		d_printf ("Blocked write queue %s\n", done_writes ?
-			  "flushed & empty" : "still active");
-
-		if (done_writes) /* drop G_IO_OUT */
-			linc_watch_set_condition (cnx->priv->tag,
-						  LINC_ERR_CONDS | LINC_IN_CONDS);
-
+		linc_connection_flush_write_queue (cnx);
 	}
 
 	if (condition & (LINC_ERR_CONDS | G_IO_OUT)) {
+		int rv, n;
+		LincSockLen n_size = sizeof (n);
+
 		switch (cnx->status) {
 		case LINC_CONNECTING:
 			n = 0;
@@ -917,7 +1017,11 @@ linc_connection_io_handler (GIOChannel  *gioc,
 					LINC_ERR_CONDS | LINC_IN_CONDS);
 
 				linc_connection_state_changed (cnx, LINC_CONNECTED);
-				
+
+				if (cnx->priv->write_queue) {
+					d_printf ("Connected, with queued writes, start flush ...\n");
+					linc_connection_flush_write_queue (cnx);
+				}
 			} else {
 				d_printf ("Error connecting %d %d %d on fd %d\n",
 					   rv, n, errno, cnx->priv->fd);
@@ -936,7 +1040,59 @@ linc_connection_io_handler (GIOChannel  *gioc,
 		}
 	}
 
-	g_object_unref (G_OBJECT (cnx));
+	CNX_UNLOCK (cnx);
+	linc_object_unref (cnx);
 
 	return TRUE;
 }
+
+LINCConnectionStatus
+linc_connection_get_status (LINCConnection *cnx)
+{
+	LINCConnectionStatus status;
+
+	CNX_LOCK (cnx);
+	status = cnx->status;
+	CNX_UNLOCK (cnx);
+
+	d_printf ("Get status on %p = %d", cnx, status);
+
+	return status;
+}
+
+void
+linc_connection_exec_disconnect (LINCCommandDisconnect *cmd)
+{
+	d_printf ("Exec defered disconnect on %p", cmd->cnx);
+
+	CNX_LOCK (cmd->cnx);
+	linc_connection_state_changed (cmd->cnx, LINC_DISCONNECTED);
+	CNX_UNLOCK (cmd->cnx);
+
+	linc_object_unref (cmd->cnx);
+	g_free (cmd);
+}
+
+void
+linc_connection_disconnect (LINCConnection *cnx)
+{
+	LINCCommandDisconnect *cmd;
+
+	cmd = g_new (LINCCommandDisconnect, 1);
+	cmd->cmd.type = LINC_COMMAND_DISCONNECT;
+	cmd->cnx = linc_object_ref (cnx);
+
+	linc_exec_command ((LINCCommand *) cmd);
+}
+
+LINCConnectionStatus
+linc_connection_wait_connected (LINCConnection *cnx)
+{
+	g_return_val_if_fail (!linc_get_threaded (),
+			      LINC_CONNECTED);
+	
+	while (cnx && cnx->status == LINC_CONNECTING) 
+		linc_main_iteration (TRUE);
+
+	return cnx ? cnx->status : LINC_DISCONNECTED;
+}
