@@ -47,12 +47,14 @@ make_local_tmpdir (const char *dirname)
 			if (stat (dirname, &statbuf) != 0)
 				g_error ("Can not stat %s\n", dirname);
 
+#ifndef __CYGWIN__
 			if (statbuf.st_uid != getuid ())
 				g_error ("Owner of %s is not the current user\n", dirname);
 
 			if ((statbuf.st_mode & (S_IRWXG|S_IRWXO)) ||
 			    !S_ISDIR (statbuf.st_mode))
 				g_error ("Wrong permissions for %s\n", dirname);
+#endif
 
 			break;
 				
@@ -110,7 +112,7 @@ linc_get_tmpdir (void)
 #define LINC_SET_SOCKADDR_LEN(saddr, len)
 #endif
 
-#if defined(AF_INET6) && defined(RES_USE_INET6)
+#if defined(HAVE_RESOLV_H) && defined(AF_INET6) && defined(RES_USE_INET6)
 #define LINC_RESOLV_SET_IPV6     _res.options |= RES_USE_INET6
 #define LINC_RESOLV_UNSET_IPV6   _res.options &= ~RES_USE_INET6
 #else
@@ -281,7 +283,7 @@ linc_protocol_is_local_ipv46 (const LINCProtocolInfo *proto,
 			struct in_addr ipv4_addr;
 			
 			if (!ipv4_addr_from_addr (&ipv4_addr,
-						  local_hostent->h_addr_list [i],
+						  (guint8 *)local_hostent->h_addr_list [i],
 						  local_hostent->h_length))
 				continue;
 
@@ -358,8 +360,10 @@ linc_protocol_get_sockaddr_ipv4 (const LINCProtocolInfo *proto,
 	        int i;
 
 		LINC_RESOLV_UNSET_IPV6;
+#ifdef HAVE_RESOLV_H
 		if (!(_res.options & RES_INIT))
 			res_init();
+#endif
 		
 		host = gethostbyname (hostname);
 		if (!host) {
@@ -452,8 +456,10 @@ linc_protocol_get_sockaddr_ipv6 (const LINCProtocolInfo *proto,
 	return (struct sockaddr *)saddr;
 #else
 
+#ifdef HAVE_RESOLV_H
 	if (!(_res.options & RES_INIT))
 		res_init();
+#endif
 
 	LINC_RESOLV_SET_IPV6;
 	host = gethostbyname (hostname);
@@ -493,7 +499,7 @@ linc_protocol_get_sockaddr_unix (const LINCProtocolInfo *proto,
 {
 	struct sockaddr_un *saddr;
 	int                 pathlen;
-	char                buf[64], *actual_path;
+	char                buf[LINC_UNIX_PATH_MAX], *actual_path;
 
 	g_assert (proto->family == AF_UNIX);
 
@@ -520,9 +526,9 @@ linc_protocol_get_sockaddr_unix (const LINCProtocolInfo *proto,
 	} else 
 		actual_path = (char *)path;
 
-	pathlen = strlen (actual_path);
+	pathlen = strlen (actual_path) + 1;
 
-	if (pathlen >= sizeof (saddr->sun_path))
+	if (pathlen > sizeof (saddr->sun_path))
 		return NULL;
 
 	saddr = g_new0 (struct sockaddr_un, 1);
@@ -694,10 +700,17 @@ linc_protocol_get_sockinfo_ipv4 (const LINCProtocolInfo  *proto,
  */
 #ifdef AF_INET6
 
+/*
+ *   We need some explicit check for Macs here - OSF1 does this
+ * right, and does not use a #define; so the Mac gets to break for
+ * now, until someone sends me a patch.
+ */
+#ifdef MAC_OS_X_IS_SO_BROKEN
 /* FIXME: is IN6ADDR_ANY_INIT exported on Mac OS X ? */
 /* on Mac OS X 10.1 inaddr6_any isn't exported by libc */
-#ifndef in6addr_any
+#  ifndef in6addr_any
 	static const struct in6_addr in6addr_any = { { { 0 } } };
+#  endif
 #endif
 
 static gboolean
@@ -707,7 +720,7 @@ linc_protocol_get_sockinfo_ipv6 (const LINCProtocolInfo  *proto,
 				 gchar                  **portnum)
 {
 	struct sockaddr_in6 *sa_in6 = (struct sockaddr_in6 *)saddr;
-#ifdef HAVE_GETADDRINFO
+#ifdef HAVE_GETNAMEINFO
 	char hbuf[NI_MAXHOST];
 #else
 	struct hostent      *host = NULL;
